/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>


extern "C" {
#include <glib.h>
}

#include <cutil.h>
#include <cufft.h>

/*
#define CUDA_SAFE_CALL(x) (x)
#define CUFFT_SAFE_CALL(x) (x)
*/

#include "debug.h"
#include "hw_tools.h"
#include "hst_setup.h"
#include "hst_reconstructor.h"

#include "hst_cuda.h"

#define MAXNPROJECTIONS 4000			//!< Maximum number of supported projections per slices
#define HST_CUDA_MAX_DEVICES 16			//!< Maximum CUDA-devices supported by the module (check tex_projes)


#define BLOCK_SIZE 16				//!< CUDA block size (for filtering)
#define BLOCK_SIZE_X BLOCK_SIZE			//!< Check kernels, there are some expectations
#define BLOCK_SIZE_Y BLOCK_SIZE			//!< Check kernels, there are some expectations

    /* We should be carefoul here, in many cases it faster to preform everything at once
    probably due to the texture caching */
#define BP_BATCH_SIZE 8				//!< Batch size (in blocks) for Back Projection 
#define FFT_BATCH_SIZE 8			//!< Batch size (in blocks) for FFT transformations (should be even)

#include "hst_cuda_kernels.h"
#include "hst_cuda_int_kernels.h"

#ifdef HST_TEXTURE16
# include <npp.h>
#endif /* HST_TEXTURE16 */


static int hst_cuda_device_num[HST_CUDA_MAX_DEVICES];
static cudaDeviceProp hst_cuda_device_prop[HST_CUDA_MAX_DEVICES];	//!< Enumerated CUDA-enabled devices
static HSTConstString hst_cuda_timers[] = { "complete", "data transfer", "filtering", "backprojection", NULL }; //!< List of supported timers

#define GPU_CONTEXT(ctx) ((GPUContext*)ctx)

/**
 * This implementation of HSTReconstructor uses NVidia GeForce-family graphic
 * cards and NVidia Tesla accelerators to accelerate reconstruction process.
 * The implementation is based on CUDA toolkit and uses NVidia cuFFT library
 * for performing Fourier Transformations.
 * GPUContext is extension of #HSTReconstructorContext which provides additional
 * data members needed to communicate with graphic hardware
 */
struct GPUContextT {
    HSTReconstructorContext recon;
#ifdef PYHST_MEASURE_TIMINGS
# define GPU_CONTEXT_MEMSET_OFFSET (3 * sizeof(GTimer*))
    GTimer *main_timer;		//!< Counts time spent in backprojection code
    GTimer *pre_timer;		//!< Counts time spent in filtering
    GTimer *mem_timer;		//!< Counts time spent in memory operations
#else
# define GPU_CONTEXT_MEMSET_OFFSET 0
#endif /* PYHST_MEASURE_TIMINGS */

    int device;			//!< Sequence number [0..MAX_DEVICES-1]
    int initialized;		//!< At least partly
    
    cudaStream_t stream[2];	//!< For interleaving memory writes and computations

    int fft_batch_size;		//!< Number of projections to process at once
    int bp_batch_size;		//!< Number of rows to process at once
    int bin_pitch_size;		//!< The number of floats to allocate for each projection (gpu_data)
    
    int bp_grid_lines;		//!< Number of CUDA blocks along Y-axis of the result slice (batched processing)
    int bp_grid_columns;	//!< Number of CUDA blocks along X-axis of the result slice 
    int projection_grid_size;	//!< Number of CUDA blocks along projections (each block consists of BLOCK_SIZE threads)
    int bin_grid_size;		//!< Number of CUDA blocks along projection bins (each block consists of BLOCK_SIZE threads)
    int filter_grid_size;	//!< Horizontal CUDA blocks for filtering (2 * dim_fft / BLOCK_SIZE)
    
#ifdef HST_TEXTURE16
    int allprojection_grid_size;//!< Number of CUDA blocks along all projections (each block consists of BLOCK_SIZE threads)
#endif /* HST_TEXTURE16 */

    int mmax_buffer_size;	//!< Buffer size required for float to uint16 conversion of projection data


    cudaChannelFormatDesc float_desc;	//!< Texture channel description
    cufftHandle fft_plan;		//!< Complex plan for fourier transformations (both forward and inverse)

    float *gpu_limits;		//!< Parameters for fai360-mode corrections
    float *gpu_data;		//!< Sinogram buffer in device memory
    float *gpu_buffer;		//!< Temporary computation buffer in GPU memory (for filtering, can hold a single batch)
    float *gpu_filter;		//!< Filter buffer in device memory
    float *gpu_result;		//!< Reconstructed slice is going here

#ifdef HST_TEXTURE16
    unsigned short *gpu_intdata;//!< Sinogram integer buffer
#endif /* HST_TEXTURE16 */
};
typedef struct GPUContextT GPUContext;


#define hst_cuda_calc_blocks hw_calc_blocks

static int hst_cuda_configure_limits(GPUContext *ctx, HSTSetup *setup) {
    float *param_s;

    int projection;
    int num_proj = setup->num_projections;
    int num_bins = setup->num_bins;
    float slope_zone = setup->pente_zone;

    float param;
    float axis_position_corr;
    float overlapping, flat_zone;

    
    param_s = (float*)malloc(2 * num_proj * sizeof(float));
    if (!param_s) return ENOMEM;

    for (projection = 0; projection < num_proj; projection++) {
	axis_position_corr = setup->axis_position_corr_s[projection];

        if (2 * axis_position_corr > num_bins) {
            overlapping = num_bins - axis_position_corr;
            param = 1; // pente_zone / prof_fact
        } else {
            overlapping = axis_position_corr;
            param = -1; // pente_zone / prof_fact
        }

        if (overlapping <= 0) return ERANGE;

        slope_zone = MIN(slope_zone, overlapping);
        flat_zone = overlapping - slope_zone;

	param_s[2*projection] = flat_zone;
	param_s[2*projection + 1] = param * slope_zone;

    }

#ifdef PYHST_MEASURE_TIMINGS
    g_timer_continue(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */
    CUDA_SAFE_CALL(cudaMemcpy(ctx->gpu_limits, param_s, 2*num_proj*sizeof(float), cudaMemcpyHostToDevice));
#ifdef PYHST_MEASURE_TIMINGS
    g_timer_stop(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    free(param_s);

    return 0;
}



/**
 *
 * Create GPU context (uninitialized)
 *
 * @param prototype is pointer on HSTReconstructure describing the reconstruction module
 * @param setup is pointer on HSTSetup with various HST parameters
 * @result created context
 */
static HSTReconstructorContext *hst_cuda_create_context(HSTReconstructor *prototype, HSTSetup *setup, int id) {
    GPUContext *ctx;
    
    assert(prototype);
    assert(setup);

    assert(id < prototype->devices);
    
    /* FIXME: no error code in case of out-of-memory */
    ctx = (GPUContext*)malloc(sizeof(struct GPUContextT));
    if (ctx) {
#ifdef PYHST_MEASURE_TIMINGS
        ctx->main_timer = g_timer_new();
        if (ctx->main_timer) g_timer_stop(ctx->main_timer);
        ctx->pre_timer = g_timer_new();
        if (ctx->pre_timer) g_timer_stop(ctx->pre_timer);
        ctx->mem_timer = g_timer_new();
        if (ctx->mem_timer) g_timer_stop(ctx->mem_timer);

        if ((!ctx->main_timer)||(!ctx->pre_timer)||(!ctx->mem_timer)) {
            if (ctx->main_timer) g_timer_destroy(ctx->main_timer);
            if (ctx->pre_timer) g_timer_destroy(ctx->pre_timer);
            if (ctx->mem_timer) g_timer_destroy(ctx->mem_timer);
            free(ctx);
            return NULL;
        }
#endif /* PYHST_MEASURE_TIMINGS */

	hst_reconstructor_init_context((HSTReconstructorContext*)ctx, prototype, setup);
    
        ctx->device = id;
    }
    return (HSTReconstructorContext*)ctx;
}

/**
  * Initializes GPU context
  *
  * @param ctx is uninitialized GPU context
  * @result return code and 0 indicates success
  */
static int hst_cuda_init_context(HSTReconstructorContext *rctx, HWThread thr) {
    GPUContext *ctx;
    HSTSetup *setup;
    
    int talign;				// The mandatory alignment of texture line in floats
    int dim_result;			// The Y-dimmension of result buffer
    int num_projections;		// Projection dimensions (allocation size, rounded for blocked processing)
    int filled_projections;		// Number of projections which will be filled in the kernels


    assert(rctx);

    ctx = GPU_CONTEXT(rctx);
    setup = rctx->setup;
    
    assert(setup);
	/* This limitation is due to statically allocated memory for sin_s, cos_s, etc. It probably could be done
	dinamically in run-time, just access performance should be checked */
    assert(setup->num_projections < MAXNPROJECTIONS);
	/* Otherwise, the filter kernel will corrupt the data and we need to limit the CUDA block_size here as well */
    assert (2 * setup->dim_fft >= BLOCK_SIZE);
    
    /* Why the width is doubled?
	We will compute two real convolutions as a single complex convolution.
	To achieve that we interleaving the data from two float arrays into 
	the complex one (i.e. first real vector is forming real part of complex
	numbers and second - the imaginary part). Then, we perform the C2C
	convolution of resulting vector and disassemble the resulting complex
	vector into the two real-values sequences.
	For details, see:
	http://www.engineeringproductivitytools.com/stuff/T0001/PT10.HTM
    */
    
    ctx->bin_grid_size = hst_cuda_calc_blocks(setup->num_bins, BLOCK_SIZE, NULL);
    ctx->bin_pitch_size = ctx->bin_grid_size * BLOCK_SIZE;

    talign = hst_cuda_device_prop[ctx->device].textureAlignment / sizeof(float);
    ctx->bin_pitch_size = talign * hst_cuda_calc_blocks(ctx->bin_pitch_size, talign, NULL);

    ctx->filter_grid_size = 2 * setup->dim_fft / BLOCK_SIZE;
    if (setup->num_projections < BLOCK_SIZE * FFT_BATCH_SIZE) {
	num_projections = setup->num_projections;
	if (num_projections%2) ++num_projections;
	
	 ctx->fft_batch_size = num_projections;
	
	ctx->projection_grid_size = hst_cuda_calc_blocks(num_projections/2, BLOCK_SIZE, NULL);
	num_projections = ctx->projection_grid_size * BLOCK_SIZE;
	
    } else { 
	ctx->fft_batch_size = BLOCK_SIZE * FFT_BATCH_SIZE;
	ctx->projection_grid_size = FFT_BATCH_SIZE / 2;
	
	num_projections = ctx->fft_batch_size * hst_cuda_calc_blocks(setup->num_projections, ctx->fft_batch_size, NULL);
    }	
    
    filled_projections = num_projections;
    num_projections = 2 * BLOCK_SIZE_Y * hst_cuda_calc_blocks(num_projections, 2 * BLOCK_SIZE_Y, NULL);

    ctx->bp_grid_columns = (int)(setup->num_x * 1.0 / BLOCK_SIZE_X + 0.9999999999);
    if ((!BP_BATCH_SIZE)||(setup->num_y < BLOCK_SIZE_Y * BP_BATCH_SIZE)) {
	ctx->bp_batch_size = setup->num_y;

	ctx->bp_grid_lines = (int)(setup->num_y * 1.0 / BLOCK_SIZE_Y + 0.9999999999);

	dim_result = ctx->bp_grid_lines * BLOCK_SIZE_Y;
    } else {
	ctx->bp_batch_size = BLOCK_SIZE_Y * BP_BATCH_SIZE;
	ctx->bp_grid_lines = BP_BATCH_SIZE;
	
	dim_result = ctx->bp_batch_size * hst_cuda_calc_blocks(setup->num_y, ctx->bp_batch_size, NULL);
    }

#ifdef HST_TEXTURE16
    ctx->allprojection_grid_size = hst_cuda_calc_blocks(setup->num_projections, BLOCK_SIZE, NULL);
#endif /* HST_TEXTURE16 */
    

    ctx->initialized = 1;    

#ifdef PYHST_MEASURE_TIMINGS
    g_timer_continue(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */

#ifdef HW_USE_THREADS
    CUDA_SAFE_CALL(cudaSetDevice(hst_cuda_device_num[ctx->device]));
#endif /* HW_USE_THREADS */
    
#ifdef HST_TEXTURE16
# ifdef HST_TEXTURE16_PACKED
    ctx->float_desc = cudaCreateChannelDesc<uint32_t>();
# else /* HST_TEXTURE16_PACKED */
    ctx->float_desc = cudaCreateChannelDesc<uint16_t>();
# endif /* HST_TEXTURE16_PACKED */

    int_projes.filterMode = cudaFilterModePoint;
    int_projes.addressMode[0] = cudaAddressModeClamp;
    int_projes.addressMode[1] = cudaAddressModeClamp;

    nppsMinMaxGetBufferSize_32f(setup->num_projections * ctx->bin_pitch_size, &ctx->mmax_buffer_size); // DS: Argument in elements, result in bytes?
    ctx->mmax_buffer_size += 2 * sizeof(float); // min & max
#else /* HST_TEXTURE16 */
    ctx->float_desc = cudaCreateChannelDesc<float>();
#endif /* HST_TEXTURE16 */

    tex_projes.filterMode = cudaFilterModeLinear; //cudaFilterModePoint;
    tex_projes.addressMode[0] = cudaAddressModeBorder;
    tex_projes.addressMode[1] = cudaAddressModeBorder;
    
    float4 *all = (float4*)malloc(setup->num_projections * sizeof(float4));
//    int max_num = 0, num;
    for (int i = 0; i < setup->num_projections; i++) {
	all[i].x = setup->cos_s[i];
	all[i].y = setup->sin_s[i];
	all[i].z = setup->axis_position_corr_s[i];
	all[i].w = floor(MIN(MIN(BLOCK_SIZE_X * all[i].x, - BLOCK_SIZE_Y * all[i].y), MIN(0., BLOCK_SIZE_X * all[i].x - BLOCK_SIZE_Y * all[i].y)));
	
//	num = ceil(MAX(MAX(BLOCK_SIZE_X * all[i].x, - BLOCK_SIZE_Y * all[i].y), MAX(0., BLOCK_SIZE_X * all[i].x - BLOCK_SIZE_Y * all[i].y))) - all[i].w;
//	if (num > max_num) max_num = num;
    }
//    printf("Max: %i\n", max_num);
    CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_all, all, setup->num_projections*sizeof(float4)));
    free(all);
    
//    CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_coss, setup->cos_s, setup->num_projections*sizeof(float)));
//    CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_sins, setup->sin_s, setup->num_projections*sizeof(float)));
//    CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_axiss, setup->axis_position_corr_s, setup->num_projections*sizeof(float)));

    CUDA_SAFE_CALL(cudaMalloc((void**)&ctx->gpu_data, num_projections*ctx->bin_pitch_size*sizeof(float)));

    if (filled_projections < num_projections) {
	CUDA_SAFE_CALL(cudaMemset(ctx->gpu_data + filled_projections * ctx->bin_pitch_size, 0, (num_projections - filled_projections)*ctx->bin_pitch_size*sizeof(float)));
    }

#ifdef HST_TEXTURE16
    CUDA_SAFE_CALL(cudaMemset(ctx->gpu_data, 0, setup->num_projections*ctx->bin_pitch_size*sizeof(float)));

    CUDA_SAFE_CALL(cudaMalloc((void**)&ctx->gpu_intdata, num_projections*ctx->bin_pitch_size*sizeof(unsigned short)));
    CUDA_SAFE_CALL(cudaMemset(ctx->gpu_intdata, 0, num_projections*ctx->bin_pitch_size*sizeof(unsigned short)));

    if (filled_projections < num_projections) {
	CUDA_SAFE_CALL(cudaMemset(ctx->gpu_intdata + filled_projections * ctx->bin_pitch_size, 0, (num_projections - filled_projections)*ctx->bin_pitch_size*sizeof(unsigned short)));
    }
#endif /* HST_TEXTURE16 */



    CUDA_SAFE_CALL(cudaMalloc((void**)&ctx->gpu_buffer, MAX(ctx->fft_batch_size*setup->dim_fft*sizeof(float), ctx->mmax_buffer_size)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&ctx->gpu_filter, 2*setup->dim_fft*sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&ctx->gpu_limits, 2*num_projections*sizeof(float)));

    CUDA_SAFE_CALL(cudaMalloc((void**)&ctx->gpu_result, dim_result * ctx->bp_grid_columns * BLOCK_SIZE_X * sizeof(float)));
//    CUDA_SAFE_CALL(cudaMemset(ctx->gpu_result, 0, sizeof(float) * dim_result));

    CUFFT_SAFE_CALL(cufftPlan1d(&ctx->fft_plan, setup->dim_fft, CUFFT_C2C, ctx->fft_batch_size/2));

    cudaStreamCreate(&ctx->stream[0]);
    cudaStreamCreate(&ctx->stream[1]);
    
    /* 
	Shall we create a smaller plan for the last iteration?
    */
    
#ifdef PYHST_MEASURE_TIMINGS
    g_timer_stop(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    return 0;
}

/**
  * Free resources occupied by GPU context
  *
  * @param ctx is initialized GPU context
  */    
static void hst_cuda_free_context(HSTReconstructorContext *rctx) {
    GPUContext *ctx;

    assert(rctx);
    ctx = GPU_CONTEXT(rctx);
    assert(ctx);

    if (ctx->initialized) {
#ifdef PYHST_MEASURE_TIMINGS
	g_timer_continue(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */
        if (ctx->stream[1]) cudaStreamDestroy(ctx->stream[1]);
	if (ctx->stream[0]) cudaStreamDestroy(ctx->stream[0]);
    
	if (ctx->fft_plan) CUFFT_SAFE_CALL(cufftDestroy(ctx->fft_plan));

	if (ctx->gpu_limits) CUDA_SAFE_CALL(cudaFree(ctx->gpu_limits));
	if (ctx->gpu_result) CUDA_SAFE_CALL(cudaFree(ctx->gpu_result));
	if (ctx->gpu_filter) CUDA_SAFE_CALL(cudaFree(ctx->gpu_filter));
	if (ctx->gpu_buffer) CUDA_SAFE_CALL(cudaFree(ctx->gpu_buffer));
#ifdef HST_TEXTURE16
	if (ctx->gpu_intdata) CUDA_SAFE_CALL(cudaFree(ctx->gpu_intdata));
#endif /* HST_TEXTURE16 */
	if (ctx->gpu_data) CUDA_SAFE_CALL(cudaFree(ctx->gpu_data));
#ifdef PYHST_MEASURE_TIMINGS
	g_timer_stop(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */
    }

    memset(((char*)ctx) + sizeof(HSTReconstructorContext) + GPU_CONTEXT_MEMSET_OFFSET, 0, sizeof(GPUContext) - sizeof(HSTReconstructorContext) - GPU_CONTEXT_MEMSET_OFFSET);
}

/**
  * Free resources and destroy GPU context
  *
  * @param ctx is initialized GPU context
  */
static void hst_cuda_destroy_context(HSTReconstructorContext *rctx) {
#ifdef PYHST_MEASURE_TIMINGS
    GPUContext *ctx;
#endif /* PYHST_MEASURE_TIMINGS */
    
    hst_cuda_free_context(rctx);
    hst_reconstructor_free_context(rctx);

#ifdef PYHST_MEASURE_TIMINGS
    ctx = GPU_CONTEXT(rctx);

    g_timer_destroy(ctx->pre_timer);
    g_timer_destroy(ctx->main_timer);
    g_timer_destroy(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    free(rctx);
}


static int hst_cuda_configure(HSTReconstructorContextPtr rctx, HSTChanged what_changed) {
    GPUContext *ctx;
    HSTSetup *setup;

    assert(rctx);

    ctx = GPU_CONTEXT(rctx);
    setup = rctx->setup;

    if (what_changed&HST_FILTER_CHANGED) {
#ifdef PYHST_MEASURE_TIMINGS
	g_timer_continue(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */
	CUDA_SAFE_CALL(cudaMemcpy(ctx->gpu_filter, setup->filter, (2 * setup->dim_fft)*sizeof(float), cudaMemcpyHostToDevice));
#ifdef PYHST_MEASURE_TIMINGS
	g_timer_stop(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */
    }
    if (what_changed&HST_LIMITS_CHANGED) {
	if (setup->fai360) {
	    check_code(hst_cuda_configure_limits(ctx, setup), "hst_cuda_configure_limit");
	}
    }

    return 0;
}

static HSTConstString hst_cuda_get_title(HSTReconstructorConstContextPtr rctx) {
    assert(rctx);

    return hst_cuda_device_prop[GPU_CONTEXT(rctx)->device].name;
}


static HSTConstString *hst_cuda_get_timers(HSTReconstructorConstContextPtr rctx, double *timers) {
#ifdef PYHST_MEASURE_TIMINGS
    assert(rctx);
    
    if (timers) {
        timers[1] = g_timer_elapsed(GPU_CONTEXT(rctx)->mem_timer, NULL);
        timers[2] = g_timer_elapsed(GPU_CONTEXT(rctx)->pre_timer, NULL);
        timers[3] = g_timer_elapsed(GPU_CONTEXT(rctx)->main_timer, NULL);
        timers[0] = timers[1] + timers[2] + timers[3];
    }
#endif /* PYHST_MEASURE_TIMINGS */

    return hst_cuda_timers;
}

static int hst_cuda_reconstruct(HSTReconstructorContext *rctx, float *SLICE, const float *SINOGRAMS) {
    int i = 0;
    
    int fermi;
    
    int batch_size;//, nextbatch;
    int batch;

    GPUContext *ctx;
    HSTSetup *setup;
    
    float *gpu_data;
    float *gpu_buffer;
    const float *data;

    int dim_fft, num_proj, num_bins;
    int num_x, num_y;
    int dimrecx;
    float axis_position;

    int mid;

    assert(rctx);
    ctx = GPU_CONTEXT(rctx);
    setup = rctx->setup;
    
    assert(ctx);
    assert(setup);

    fermi = (hst_cuda_device_prop[ctx->device].major > 1);

    gpu_buffer = ctx->gpu_buffer;

    dim_fft = setup->dim_fft;
    num_proj = setup->num_projections;
    num_bins = setup->num_bins;
    num_x = setup->num_x;
    num_y = setup->num_y;
    axis_position = setup->axis_position;


    mid = (num_bins + dim_fft) / 2;

    dim3 dimPadGrid(setup->dim_fft / BLOCK_SIZE,  ctx->projection_grid_size);
    dim3 dimInputGrid(ctx->bin_grid_size, ctx->projection_grid_size);
    dim3 dimFilterGrid(ctx->filter_grid_size, ctx->projection_grid_size);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);


    batch_size = ctx->fft_batch_size;	// Could be shorter on the last iteration

/*
#ifdef PYHST_MEASURE_TIMINGS
    g_timer_continue(ctx->mem_timer);
#endif 
    CUDA_SAFE_CALL(cudaMemcpy2DAsync(ctx->gpu_data, ctx->bin_pitch_size * sizeof(float), SINOGRAMS, num_bins * sizeof(float), num_bins * sizeof(float), batch_size, cudaMemcpyHostToDevice, ctx->stream[0]));
#ifdef PYHST_MEASURE_TIMINGS
    g_timer_stop(ctx->mem_timer);
#endif 
*/

    for (batch = 0, i = 0; batch < num_proj; batch += batch_size, ++i) {
	data = SINOGRAMS + batch * num_bins;
        gpu_data = ctx->gpu_data + batch * ctx->bin_pitch_size;

	if (batch + batch_size >= num_proj) {
	    batch_size = num_proj - batch;
	}

#ifdef PYHST_MEASURE_TIMINGS
	g_timer_continue(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */
	CUDA_SAFE_CALL(cudaMemcpy2D(gpu_data, ctx->bin_pitch_size * sizeof(float), data, num_bins * sizeof(float), num_bins * sizeof(float), batch_size, cudaMemcpyHostToDevice));
/*
	if (batch + batch_size < num_proj) {
	    if (batch + 2*batch_size < num_proj) nextbatch = batch_size;
	    else nextbatch = num_proj - (batch + 2*batch_size);
	    CUDA_SAFE_CALL(cudaMemcpy2DAsync(gpu_data + batch_size*ctx->bin_pitch_size, ctx->bin_pitch_size * sizeof(float), data + batch_size*num_bins, num_bins * sizeof(float), num_bins * sizeof(float), nextbatch, cudaMemcpyHostToDevice, ctx->stream[(i+1)%2]));
	}
	cudaStreamSynchronize(ctx->stream[i%2]);
*/
#ifdef PYHST_MEASURE_TIMINGS
	g_timer_stop(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */



#ifdef PYHST_MEASURE_TIMINGS
	g_timer_continue(ctx->pre_timer);
#endif /* PYHST_MEASURE_TIMINGS */

        if (dim_fft > num_bins) {
            if (setup->zero_padding) {
		hst_cuda_pack_kernel_zpad<<<dimPadGrid,dimBlock>>>(gpu_buffer, dim_fft, gpu_data, ctx->bin_pitch_size, num_bins);
	    } else {
		hst_cuda_pack_kernel_epad<<<dimPadGrid,dimBlock>>>(gpu_buffer, dim_fft, gpu_data, ctx->bin_pitch_size, num_bins, mid);
	    }
	} else {
	    hst_cuda_pack_kernel<<<dimInputGrid,dimBlock>>>(gpu_buffer, dim_fft, gpu_data, ctx->bin_pitch_size);
	}

        CUFFT_SAFE_CALL(cufftExecC2C(ctx->fft_plan, (cufftComplex*)gpu_buffer, (cufftComplex*)gpu_buffer, CUFFT_FORWARD));
        hst_cuda_filter_kernel<<<dimFilterGrid,dimBlock>>>(2*dim_fft, gpu_buffer, ctx->gpu_filter);
        CUFFT_SAFE_CALL(cufftExecC2C(ctx->fft_plan, (cufftComplex*)gpu_buffer, (cufftComplex*)gpu_buffer, CUFFT_INVERSE));

	
        if (setup->fai360) {
            hst_cuda_unpack_kernel_fai360<<<dimInputGrid,dimBlock>>>(gpu_data, ctx->bin_pitch_size, gpu_buffer, dim_fft, ctx->bin_grid_size * BLOCK_SIZE / 2, ctx->gpu_limits + 2 * batch, batch);
	} else {
 	    hst_cuda_unpack_kernel<<<dimInputGrid,dimBlock>>>(gpu_data, ctx->bin_pitch_size, gpu_buffer, dim_fft);
	}
	
#ifdef PYHST_MEASURE_TIMINGS
	g_timer_stop(ctx->pre_timer);
#endif /* PYHST_MEASURE_TIMINGS */
    }	

#ifdef PYHST_MEASURE_TIMINGS
    g_timer_continue(ctx->main_timer);
#endif /* PYHST_MEASURE_TIMINGS */

#ifdef HST_TEXTURE16
    nppsMinMax_32f(ctx->gpu_data, num_proj * ctx->bin_pitch_size, (float*)(((char*)ctx->gpu_buffer) + ctx->mmax_buffer_size), (float*)(((char*)ctx->gpu_buffer) + ctx->mmax_buffer_size + sizeof(float)), (Npp8u*)ctx->gpu_buffer);

    dim3 dimAllInputGrid(ctx->bin_grid_size, ctx->allprojection_grid_size);
    hst_cuda_normalize_kernel<<<dimAllInputGrid,dimBlock>>>(ctx->gpu_intdata, ctx->gpu_data, ctx->bin_pitch_size, (float*)(((char*)ctx->gpu_buffer) + ctx->mmax_buffer_size));
# ifdef HST_TEXTURE16_PACKED
    CUDA_SAFE_CALL( cudaBindTexture2D(NULL, int_projes, ctx->gpu_intdata, ctx->float_desc, num_bins/2 + num_bins%2, num_proj, ctx->bin_pitch_size * sizeof(unsigned short)) );
# else /* HST_TEXTURE16_PACKED */
    CUDA_SAFE_CALL( cudaBindTexture2D(NULL, int_projes, ctx->gpu_intdata, ctx->float_desc, num_bins, num_proj, ctx->bin_pitch_size * sizeof(unsigned short)) );
# endif /* HST_TEXTURE16_PACKED */
#else /* HST_TEXTURE16 */
    CUDA_SAFE_CALL( cudaBindTexture2D(NULL, tex_projes, ctx->gpu_data, ctx->float_desc, num_bins, num_proj, ctx->bin_pitch_size * sizeof(float)) );
#endif /* HST_TEXTURE16 */

#ifdef PYHST_MEASURE_TIMINGS
    g_timer_stop(ctx->main_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    dim3 dimBPBlock(BLOCK_SIZE_X, BLOCK_SIZE_Y);
    dim3 dimGrid(ctx->bp_grid_columns, ctx->bp_grid_lines );
    dimrecx = ctx->bp_grid_columns * BLOCK_SIZE_X;

    batch_size = ctx->bp_batch_size;	// Could be shorter on the last iteration

    for (batch = 0, i = 0; batch < num_y; batch += batch_size, ++i) {
	if (batch + batch_size >= num_y) {
	    batch_size = num_y - batch;
	}

#ifdef PYHST_MEASURE_TIMINGS
	g_timer_continue(ctx->main_timer);
#endif /* PYHST_MEASURE_TIMINGS */
	
#ifdef HST_TEXTURE16
# ifdef HST_TEXTURE16_PACKED
        hst_cuda_intkernel<<<dimGrid, dimBPBlock, 0, ctx->stream[i%2]>>> (num_proj, num_bins, (uint32_t*)ctx->gpu_result, setup->offset_x  -  axis_position - 0.5f, setup->offset_y - axis_position - 0.5f, batch);
# else /* HST_TEXTURE16_PACKED */
        hst_cuda_intkernel_esrf<<<dimGrid, dimBPBlock, 0, ctx->stream[i%2]>>> (num_proj, num_bins, (uint32_t*)ctx->gpu_result, setup->offset_x  -  axis_position - 0.5f, setup->offset_y - axis_position - 0.5f, batch);
# endif /* HST_TEXTURE16_PACKED */
	hst_cuda_denormalize_kernel<<<dimGrid,dimBPBlock, 0, ctx->stream[i%2]>>>(ctx->gpu_result + dimrecx * batch, (float*)(((char*)ctx->gpu_buffer) + ctx->mmax_buffer_size), num_proj);
#else /* HST_TEXTURE16 */
	if (fermi) {
# ifndef HST_FAST_MODE
	    if (setup->oversampling == 1) {
# endif /* ! HST_FAST_MODE */
    		hst_cuda_nearest_kernel<<<dimGrid, dimBPBlock, 0, ctx->stream[i%2]>>> (num_proj, num_bins, ctx->gpu_result, setup->offset_x  -  axis_position - 0.5f, setup->offset_y - axis_position - 0.5f, batch);
# ifndef HST_FAST_MODE
	    } else if (setup->oversampling) {
    		hst_cuda_oversample4_kernel<<<dimGrid, dimBPBlock, 0, ctx->stream[i%2]>>> (num_proj, num_bins, ctx->gpu_result, setup->offset_x  -  axis_position - 0.5f, setup->offset_y - axis_position - 0.5f, batch);
	    } else {
    		hst_cuda_linear_kernel<<<dimGrid, dimBPBlock, 0, ctx->stream[i%2]>>> (num_proj, num_bins, ctx->gpu_result, setup->offset_x  -  axis_position - 0.5f, setup->offset_y - axis_position - 0.5f, batch);
	    }
	} else {
	    hst_cuda_kernel<<<dimGrid, dimBPBlock, 0, ctx->stream[i%2]>>> (num_proj, num_bins, ctx->gpu_result, setup->offset_x  -  axis_position - 0.5f, setup->offset_y - axis_position - 0.5f, batch);
	}
# endif /* ! HST_FAST_MODE */
#endif /* HST_TEXTURE16 */
	
#ifdef PYHST_MEASURE_TIMINGS
	g_timer_stop(ctx->main_timer);
	g_timer_continue(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */

        CUDA_SAFE_CALL(cudaMemcpy2DAsync(SLICE + num_x * batch, num_x*sizeof(float), ctx->gpu_result + dimrecx * batch, dimrecx * sizeof(float), num_x * sizeof(float), batch_size, cudaMemcpyDeviceToHost, ctx->stream[i%2]));

#ifdef PYHST_MEASURE_TIMINGS
	g_timer_stop(ctx->mem_timer);
#endif /* PYHST_MEASURE_TIMINGS */
    }

#ifdef PYHST_MEASURE_TIMINGS
    g_timer_continue(ctx->main_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    cudaUnbindTexture(tex_projes);
    cudaThreadSynchronize();
    
#ifdef PYHST_MEASURE_TIMINGS
    g_timer_stop(ctx->main_timer);
#endif /* PYHST_MEASURE_TIMINGS */

    return 0;
}



static HSTReconstructor hst_gpu_info = {
    0,
    hst_cuda_get_title,
    hst_cuda_create_context,
    hst_cuda_init_context,
    hst_cuda_free_context,
    hst_cuda_destroy_context,
    hst_cuda_configure,
    NULL,
    hst_cuda_reconstruct,
    hst_reconstructor_postprocess_slice,
    hst_cuda_get_timers
};


HSTReconstructor *hst_cuda_init(int flags) {
    int i, dev;
    int device_count;

    CUDA_SAFE_CALL(cudaGetDeviceCount(&device_count));
    
    if (device_count > HST_CUDA_MAX_DEVICES) {
	pyhst_warning("There is %u CUDA-enabled devices detected in the system, but hst_cuda is configured to use only %u", device_count, HST_CUDA_MAX_DEVICES);
	device_count = HST_CUDA_MAX_DEVICES;
    }

    pyhst_info("NVIDIA devices: %u", device_count);
    for (i = 0, dev = 0; i < device_count; ++i) {
        CUDA_SAFE_CALL(cudaGetDeviceProperties(&hst_cuda_device_prop[dev], i));

        if ((hst_cuda_device_prop[dev].computeMode == cudaComputeModeProhibited)||((hst_cuda_device_prop[dev].major == 9999 && hst_cuda_device_prop[dev].minor == 9999))) continue;
#ifdef HW_IGNORE_OLD_HARDWARE
	if (hst_cuda_device_prop[dev].major < 2) continue;
#endif /* HW_IGNORE_OLD_HARDWARE */
        pyhst_debug(" * Device %d: %s v. %d.%d: %d SM, %.2f MHZ, %d MB%s", dev, hst_cuda_device_prop[dev].name, hst_cuda_device_prop[dev].major, hst_cuda_device_prop[dev].minor, hst_cuda_device_prop[dev].multiProcessorCount, hst_cuda_device_prop[dev].clockRate * 1e-6, hst_cuda_device_prop[dev].totalGlobalMem/1048576, hst_cuda_device_prop[dev].deviceOverlap?", overlap":"");

        hst_cuda_device_num[dev++] = i;
    }


    hst_gpu_info.devices = dev;
    return &hst_gpu_info;
}

void hst_cuda_free() {
}

void *hst_cuda_host_malloc(size_t size) {
    cudaError_t err;
    void *tmp;

    if (hst_gpu_info.devices) {
	err = cudaHostAlloc(&tmp, size, cudaHostAllocPortable);//4);//cudaHostAllocWriteCombined);
	if (err) return NULL;
    } else {
	tmp = malloc(size);
    }
    return tmp;
}

void hst_cuda_host_free(void *ptr) {
    if (hst_gpu_info.devices) {
	cudaFreeHost(ptr);
    } else {
	free(ptr);
    }
}
