/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <math.h>
#include <CL/cl.h>

#include <clFFT.h>

#include <glib.h>

#include "hw_tools.h"
#include "debug.h"
#include "hst_setup.h"
#include "hst_reconstructor.h"
#include "hst_opencl_extensions.h"

#define BLOCK_SIZE 16		// Fix in kernels as well

#define HST_CHECK_MEMORY 512	// minimal size in megabytes
#define HST_OPENCL_CONST_HACK
#define HST_OPENCL_BUILD_FLAGS NULL //"-cl-denorms-are-zero -cl-mad-enable -cl-no-signed-zeros -cl-unsafe-math-optimizations -cl-finite-math-only"
//#define HST_OPENCL_LOAD_PTX "/tmp/hst_opencl.ptx"
//#define HST_OPENCL_SAVE_PTX "/tmp/hst_opencl.ptx"
#define MAX_PROJECTIONS 4000
#define HST_MAX_NAME_LENGTH 64

#define HST_MAX_OPENCL_PLATFORMS 4
#define HST_MAX_OPENCL_DEVICES 16


#ifdef HST_OPENCL_LOAD_PTX
# undef HST_OPENCL_CONST_HACK
#endif /* HST_OPENCL_LOAD_PTX */

#include "hst_opencl.h"

static int have_opencl11 = 1;

static int amd_platform = 0;
static int nvidia_platform = 0;
static int intel_platform = 0;

typedef enum {
    HST_DEVICE_UNKNOWN,
    HST_DEVICE_GPU,
    HST_DEVICE_GT200,
    HST_DEVICE_FERMI,
    HST_DEVICE_CYPRESS,
    HST_DEVICE_AMD_CPU,
    HST_DEVICE_INTEL_CPU,
    HST_DEVICE_NUMBER
} hst_device_type;


typedef enum {
    HST_KERNEL_FILTER,
    HST_KERNEL_BACKPROJECT,
    HST_KERNEL_SPREAD,
    HST_KERNEL_PACK,
    HST_KERNEL_NORMALIZE,
    HST_KERNEL_NUMBER
} hst_kernel_type;


struct OpenCLContextT {
    /* This is one ugly hack... */
    HSTReconstructorContext hst_ctx;

    int initialized;

    char                *device_name;
    hst_device_type	arch;
    cl_device_type	device_type;
    cl_device_id        device_id;
    cl_context          clctx;
    cl_command_queue    queue;
    cl_program          program;
    cl_image_format     image_format;       /* TODO: We might drop this */
    cl_kernel           kernel[HST_KERNEL_NUMBER];

    clFFT_Plan          fft_plan;

    cl_mem unified_table;   /* everything in one place */
    cl_mem filter;          /* filter coefficients */
    cl_mem slice;           /* final result buffer */
    cl_mem sinogram;
    cl_mem texture;
    cl_mem buffer;          /* keeps FFT results */
    cl_mem frequencies;     /* keeps FFT results */

    size_t bp_dim[3];
    size_t input_dim[3];
    size_t filter_dim[3];

    float off_x;
    float off_y;
    
#ifdef PYHST_MEASURE_TIMINGS
    cl_ulong timer_transfer;
    cl_ulong timer_filter;
    cl_ulong timer_backprojection;
#endif
};

typedef struct OpenCLContextT OpenCLContext;

#include "hst_opencl_error.h"
#include "hst_opencl_kernel.h"
#include "hst_opencl_timers.h"


static int hst_cuda_configure_limits(OpenCLContext *ctx, HSTSetup *setup)
{
    float *param_s;

    int projection;
    int num_proj = setup->num_projections;
    int num_bins = setup->num_bins;
    float slope_zone = setup->pente_zone;

    float param;
    float axis_position_corr;
    float overlapping, flat_zone;

    param_s = (float*) malloc(2 * num_proj * sizeof(float));
    if (!param_s)
        return ENOMEM;

    for (projection = 0; projection < num_proj; projection++) {
        axis_position_corr = setup->axis_position_corr_s[projection];

        if (2 * axis_position_corr > num_bins) {
            overlapping = num_bins - axis_position_corr;
            param = 1; // pente_zone / prof_fact
        }
        else {
            overlapping = axis_position_corr;
            param = -1; // pente_zone / prof_fact
        }

        if (overlapping <= 0)
            return ERANGE;

        slope_zone = MIN(slope_zone, overlapping);
        flat_zone = overlapping - slope_zone;

        param_s[2*projection] = flat_zone;
        param_s[2*projection + 1] = param * slope_zone;
    }

    free(param_s);

    return 0;
}

/*
 * Implementation of HSTReconstructorContext interface
 */
static HSTReconstructorContext *hst_opencl_create_context(HSTReconstructor *prototype, HSTSetup *setup, int id)
{
    OpenCLContext *ctx;

    assert(prototype);
    assert(setup);
    assert(id < prototype->devices);

    ctx = (OpenCLContext*) malloc(sizeof(struct OpenCLContextT));
    if (ctx) {
	memset(ctx, 0, sizeof(struct OpenCLContextT));
	
        hst_reconstructor_init_context((HSTReconstructorContext*) ctx, prototype, setup);
        ctx->device_id = g_hst_platform.device_ids[id];
    }
    return (HSTReconstructorContext*) ctx;
}

static int hst_opencl_init_context(HSTReconstructorContext *rctx, HWThread thrd)
{
    int err;
    int i;
    OpenCLContext *ctx;
    HSTSetup *setup;


    cl_kernel kernel;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;
    setup = rctx->setup;

    assert(setup);
    assert(setup->num_projections < MAX_PROJECTIONS);

    int num_proj = setup->num_projections;
    int num_bins = setup->num_bins;
    int num_x = setup->num_x;
    int num_y = setup->num_y;
    int dim_fft = setup->dim_fft;
    int dim_proj;

    int variants[HST_MAX_KERNEL_VARIANTS] = { 0 };


    err = clGetDeviceInfo(ctx->device_id, CL_DEVICE_TYPE, sizeof(cl_device_type), &ctx->device_type, NULL);
    CL_CHECK_ERROR(err);

    ctx->device_name = (char *) malloc(HST_MAX_NAME_LENGTH + 1);
    if (!ctx->device_name) return -1;
    
    err = clGetDeviceInfo(ctx->device_id, CL_DEVICE_NAME, HST_MAX_NAME_LENGTH, ctx->device_name, NULL);
    if (err) snprintf(ctx->device_name, 256, "OpenCL %p", ctx->device_id);

    err = clGetDeviceInfo(ctx->device_id, CL_DEVICE_NAME, HST_MAX_NAME_LENGTH, ctx->device_name, NULL);

    if (nvidia_platform) {
	cl_uint compute_major;
	err = clGetDeviceInfo(ctx->device_id, CL_DEVICE_COMPUTE_CAPABILITY_MAJOR_NV, sizeof(cl_uint), &compute_major, NULL); 
	if (err) ctx->arch = HST_DEVICE_FERMI;
	else if (compute_major < 2) ctx->arch = HST_DEVICE_GT200;
	else ctx->arch = HST_DEVICE_FERMI;
    } else if (amd_platform) {
	if (ctx->device_type == CL_DEVICE_TYPE_CPU) ctx->arch = HST_DEVICE_AMD_CPU;
	else ctx->arch = HST_DEVICE_CYPRESS;
    } else if (intel_platform) {
	ctx->arch = HST_DEVICE_INTEL_CPU;
    } else ctx->arch = HST_DEVICE_UNKNOWN;
    
    ctx->input_dim[0] = dim_fft;
    ctx->input_dim[1] = num_proj;
    ctx->filter_dim[0] = 2*dim_fft;
    ctx->filter_dim[1] = num_proj;
    ctx->bp_dim[0] = BLOCK_SIZE * hw_calc_blocks(num_x, BLOCK_SIZE, NULL);
    ctx->bp_dim[1] = BLOCK_SIZE * hw_calc_blocks(num_y, BLOCK_SIZE, NULL);

    dim_proj = 2 * BLOCK_SIZE * hw_calc_blocks(num_proj, 2 * BLOCK_SIZE, NULL);

    /* OpenCL context initialization */
    ctx->clctx = clCreateContext(0, 1, &ctx->device_id, NULL, NULL, &err);
    CL_CHECK_ERROR(err);
#ifdef PYHST_MEASURE_TIMINGS
    ctx->queue = clCreateCommandQueue(ctx->clctx, ctx->device_id, CL_QUEUE_PROFILING_ENABLE, &err);
#else
    ctx->queue = clCreateCommandQueue(ctx->clctx, ctx->device_id, 0, &err);
#endif
    CL_CHECK_ERROR(err);


    /* Allocate and copy buffers, we don't use immediate copying during buffer
     * creation, since we might copy the data to all devices. */
    cl_event event;

#ifdef HST_OPENCL_CONST_HACK
    if (ctx->arch != HST_DEVICE_GT200) {
#endif /* HST_OPENCL_CONST_HACK */
    cl_float4 *all = (cl_float4*)malloc(dim_proj * sizeof(cl_float4));
    for (i = 0; i < num_proj; i++) {
	all[i].x = setup->cos_s[i];
	all[i].y = setup->sin_s[i];
	all[i].z = setup->axis_position_corr_s[i];
	all[i].w = floor(MIN(MIN(BLOCK_SIZE * all[i].x, - BLOCK_SIZE * all[i].y), MIN(0., BLOCK_SIZE * all[i].x - BLOCK_SIZE * all[i].y)));
    }
    if (num_proj < dim_proj) memset(all + num_proj, 0, (dim_proj - num_proj) * sizeof(cl_float4));
    ctx->unified_table = clCreateBuffer(ctx->clctx, CL_MEM_READ_ONLY, dim_proj * sizeof(cl_float4), NULL, &err);
    CL_CHECK_ERROR(err);
    CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->unified_table, CL_TRUE, 0, dim_proj*sizeof(cl_float4), all, 0, NULL, &event));
    free(all);
    hst_measure_time(&ctx->timer_transfer, ctx->queue, event);
#ifdef HST_OPENCL_CONST_HACK
    }
#endif /* HST_OPENCL_CONST_HACK */

    ctx->filter = clCreateBuffer(ctx->clctx, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 2*setup->dim_fft*sizeof(float), setup->filter, &err);
    CL_CHECK_ERROR(err);

    ctx->buffer = clCreateBuffer(ctx->clctx, CL_MEM_READ_WRITE, 2*setup->dim_fft*num_proj*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

    ctx->slice = clCreateBuffer(ctx->clctx, CL_MEM_WRITE_ONLY, ctx->bp_dim[0]*ctx->bp_dim[1]*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);
    
    ctx->sinogram = clCreateBuffer(ctx->clctx, CL_MEM_READ_WRITE, num_bins*num_proj*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

	// Texture
    if (ctx->device_type == CL_DEVICE_TYPE_GPU) {
	cl_image_format volume_format;
	volume_format.image_channel_order = CL_R;
	volume_format.image_channel_data_type = CL_FLOAT;
	ctx->texture = clCreateImage2D(ctx->clctx, CL_MEM_READ_ONLY, &volume_format, num_bins, num_proj, 0, NULL, &err);
	CL_CHECK_ERROR(err);
    }

    /* Create FFT plan */
    clFFT_Dim3 dim;
    dim.x = setup->dim_fft; dim.y = 1; dim.z = 1;
    ctx->fft_plan = clFFT_CreatePlan(ctx->clctx, dim, clFFT_1D, clFFT_InterleavedComplexFormat, &err);
    CL_CHECK_ERROR(err);


#ifdef HST_FAST_MODE
    variants[HST_KERNEL_BACKPROJECT] = 1;
#else /* HST_FAST_MODE */
    variants[HST_KERNEL_BACKPROJECT] = setup->oversampling;
#endif /* HST_FAST_MODE */

    hst_opencl_load_kernels(ctx, variants);
    
    /* Set the arguments here, to avoid further calls */


    kernel = ctx->kernel[HST_KERNEL_SPREAD];
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *) &ctx->sinogram));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 2, sizeof(int), &setup->num_bins));

    kernel = ctx->kernel[HST_KERNEL_FILTER];
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *) &ctx->filter));

    kernel = ctx->kernel[HST_KERNEL_PACK];
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *) &ctx->sinogram));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 2, sizeof(int), &setup->num_bins));

    kernel = ctx->kernel[HST_KERNEL_BACKPROJECT];

    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(int), &setup->num_projections));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(int), &setup->num_bins));
    if (ctx->device_type == CL_DEVICE_TYPE_GPU) { 
	CL_CHECK_ERROR(clSetKernelArg(kernel, 4, sizeof(cl_mem), (void *) &ctx->texture)); 
    } else { 
	CL_CHECK_ERROR(clSetKernelArg(kernel, 4, sizeof(cl_mem), (void *) &ctx->sinogram));
    }
    CL_CHECK_ERROR(clSetKernelArg(kernel, 5, sizeof(cl_mem), (void *) &ctx->slice));

#ifdef HST_OPENCL_CONST_HACK
    if (ctx->arch != HST_DEVICE_GT200) {
#endif /* HST_OPENCL_CONST_HACK */
    CL_CHECK_ERROR(clSetKernelArg(kernel, 6, sizeof(cl_mem), (void *) &ctx->unified_table));
#ifdef HST_OPENCL_CONST_HACK
    }
#endif /* HST_OPENCL_CONST_HACK */

    return 0;
}

static int hst_opencl_configure(HSTReconstructorContext *rctx, HSTChanged changes)
{
    OpenCLContext *ctx;
    HSTSetup *setup;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;
    setup = rctx->setup;

    if (changes & HST_FILTER_CHANGED) {
        cl_event event;
        CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->filter, CL_TRUE, 0, 2*setup->dim_fft*sizeof(float), setup->filter, 0, NULL, &event));
        hst_measure_time(&ctx->timer_transfer, ctx->queue, event);
    }

    if (changes & HST_LIMITS_CHANGED) {
        if (setup->fai360) {
            /* FIXME: check some obscure limits */
        }
        cl_kernel kernel = ctx->kernel[HST_KERNEL_BACKPROJECT];

        ctx->off_x = setup->offset_x - setup->axis_position - 0.5f;
	ctx->off_y = setup->offset_y - setup->axis_position - 0.5f;
	CL_CHECK_ERROR(clSetKernelArg(kernel, 2, sizeof(float), &ctx->off_x));
        CL_CHECK_ERROR(clSetKernelArg(kernel, 3, sizeof(float), &ctx->off_y));
    }
    return 0;
}

static void hst_opencl_free_context(HSTReconstructorContext *rctx)
{
    hst_kernel_type kernel_id;
    OpenCLContext *ctx;
    
    assert(rctx);
    ctx = (OpenCLContext*) rctx;

    if (ctx->initialized) {
        free(ctx->device_name);
        clFFT_DestroyPlan(ctx->fft_plan);
        if (ctx->device_type == CL_DEVICE_TYPE_GPU) {
	    CL_CHECK_ERROR(clReleaseMemObject(ctx->texture));
	}
#ifdef HST_OPENCL_CONST_HACK
	if (ctx->arch != HST_DEVICE_GT200) {
#endif /* HST_OPENCL_CONST_HACK */
        CL_CHECK_ERROR(clReleaseMemObject(ctx->unified_table));
#ifdef HST_OPENCL_CONST_HACK
	}
#endif /* HST_OPENCL_CONST_HACK */

        CL_CHECK_ERROR(clReleaseMemObject(ctx->filter));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->slice));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->sinogram));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->buffer));
        CL_CHECK_ERROR(clReleaseCommandQueue(ctx->queue));
        for (kernel_id = 0; kernel_id < HST_KERNEL_NUMBER; kernel_id++) {
            CL_CHECK_ERROR(clReleaseKernel(ctx->kernel[kernel_id]));
        }
        CL_CHECK_ERROR(clReleaseProgram(ctx->program));
        CL_CHECK_ERROR(clReleaseContext(ctx->clctx));
    }

    memset(((char*) ctx) + sizeof(HSTReconstructorContext), 0, sizeof(OpenCLContext) - sizeof(HSTReconstructorContext));
}

static void hst_opencl_destroy_context(HSTReconstructorContext *rctx)
{
    hst_opencl_free_context(rctx);
    hst_reconstructor_free_context(rctx);
    free(rctx);
}

static HSTConstString hst_opencl_get_title(const HSTReconstructorContext *rctx)
{
    assert(rctx);
    
    return ((OpenCLContext *) rctx)->device_name;
}

static int hst_opencl_reconstruct(HSTReconstructorContext *rctx, float* slice, const float *sinogram)
{
    int i;
    
    OpenCLContext *ctx;
    HSTSetup *setup;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;
    setup = rctx->setup;
    assert(setup);
    assert(slice);

    int num_bins = setup->num_bins;
    int num_proj = setup->num_projections;
    int num_x = setup->num_x;
    int num_y = setup->num_y;
    cl_event event;

    cl_int err = CL_SUCCESS;

    size_t zero_offset[] = {0, 0, 0};
    size_t slice_size[] = {num_x * sizeof(float), num_y, 1};
    size_t projection_region[] = {num_bins, num_proj, 1};
    
    size_t local_work_size[] = { BLOCK_SIZE,  BLOCK_SIZE };

    if (setup->fai360) {
        /* TODO: implement */
        printf("FAI360 requested\n");
    }

    /* 1. Copy sinogram data onto device */
    CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->sinogram, CL_TRUE, 0, num_bins*num_proj*sizeof(float), sinogram, 0, NULL, &event));
    hst_measure_time(&ctx->timer_transfer, ctx->queue, event);

	
    if (!intel_platform) {	// FFT failing at the moment
    /* 2. Do FFT */
    if (setup->zero_padding) {
            /* TODO: implement zero-padding
             *
             * However, we are basically doing a zero-padding in the
             * else-branch.
             */
    } else {
        CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, ctx->kernel[HST_KERNEL_SPREAD], 2, NULL, ctx->input_dim, NULL, 0, NULL, &event));
        hst_measure_time(&ctx->timer_filter, ctx->queue, event);
    }

    CL_CHECK_ERROR(clFFT_ExecuteInterleaved(ctx->queue, ctx->fft_plan, num_proj, clFFT_Forward, ctx->buffer, ctx->buffer, 0, NULL, &event));
    hst_measure_time(&ctx->timer_filter, ctx->queue, event);

    /* 3. Filter */
    CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, ctx->kernel[HST_KERNEL_FILTER], 2, NULL, ctx->filter_dim, NULL, 0, NULL, &event));
    hst_measure_time(&ctx->timer_filter, ctx->queue, event);

    /* 4. Do inverse FFT */
    CL_CHECK_ERROR(clFFT_ExecuteInterleaved(ctx->queue, ctx->fft_plan, num_proj, clFFT_Inverse, ctx->buffer, ctx->buffer, 0, NULL, &event));
    hst_measure_time(&ctx->timer_filter, ctx->queue, event);

    /*
    kernel = ctx->kernel[HST_KERNEL_NORMALIZE];
    global_work_size[0] = dim_fft*num_proj;
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, kernel, 1, NULL, global_work_size, NULL, 0, NULL, NULL));
    */

    CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, ctx->kernel[HST_KERNEL_PACK], 2, NULL, ctx->input_dim, NULL, 0, NULL, &event));
    hst_measure_time(&ctx->timer_filter, ctx->queue, event);
    
    }

    if (ctx->device_type == CL_DEVICE_TYPE_GPU) {
	clEnqueueCopyBufferToImage(ctx->queue, ctx->sinogram, ctx->texture, 0, zero_offset, projection_region, 0, NULL, &event);
	hst_measure_time(&ctx->timer_backprojection, ctx->queue, event);
    }

    /* 5. Back-project */
    CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, ctx->kernel[HST_KERNEL_BACKPROJECT], 2, NULL, ctx->bp_dim, local_work_size, 0, NULL, &event));
    hst_measure_time(&ctx->timer_backprojection, ctx->queue, event);

/*
    GTimer *tm = g_timer_new();
    g_timer_stop(tm);
    printf("%lf\n", g_timer_elapsed(tm, NULL));
    g_timer_destroy(tm);
*/

    /* 6. Put result back into slice buffer */
//    CL_CHECK_ERROR(clEnqueueReadBuffer(ctx->queue, ctx->slice, CL_TRUE, 0, num_x*num_y*sizeof(float), slice, 0, NULL, &event));

    if (have_opencl11) {
	CL_CHECK_ERROR(clEnqueueReadBufferRect(ctx->queue, ctx->slice, CL_TRUE, zero_offset, zero_offset, slice_size, ctx->bp_dim[0] * sizeof(float), 0, 0, 0, slice, 0, NULL, &event));
    } else {
	for (i = 0; i < num_y; i++) {
	    CL_CHECK_ERROR(clEnqueueReadBuffer(ctx->queue, ctx->slice, CL_FALSE, i * ctx->bp_dim[0] * sizeof(float), num_x * sizeof(float), slice + i * num_x, 0, NULL, &event));
	}
    }
    hst_measure_time(&ctx->timer_transfer, ctx->queue, event);

    return 0;   /* XXX: What does the return value do? */
}

static HSTReconstructor hst_opencl_info = {
    0,
    hst_opencl_get_title,
    hst_opencl_create_context,
    hst_opencl_init_context,
    hst_opencl_free_context,
    hst_opencl_destroy_context,
    hst_opencl_configure,
    NULL,   /* no pre-processing */
    hst_opencl_reconstruct,
    hst_reconstructor_postprocess_slice,
    hst_opencl_get_timers
};

/*
 * Implementation of public interface
 */
HSTReconstructor *hst_opencl_init(int flags)
{
    cl_int err;
    cl_uint max = 0;
    cl_platform_id selected_platform;
    cl_uint i, num_platforms, num_devices;
    cl_platform_id platforms[HST_MAX_OPENCL_PLATFORMS];
    
    char name[128], version[128];
    
    CL_CHECK_ERROR(clGetPlatformIDs(HST_MAX_OPENCL_PLATFORMS, platforms, &num_platforms));

    if (ds_log_check_level(G_LOG_LEVEL_INFO)) {
	cl_uint num_cpu = 0, num_gpu = 0, num_accell = 0;

	pyhst_info("OpenCL Platforms: %u", num_platforms);
	for (i = 0; i < num_platforms; i++) {
    	    err = clGetPlatformInfo (platforms[i], CL_PLATFORM_NAME, sizeof(name), name, NULL);
	    if (err == CL_SUCCESS) err = clGetPlatformInfo (platforms[i], CL_PLATFORM_VERSION, sizeof(version), version, NULL);
	    if (err == CL_SUCCESS) clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_CPU, 0, NULL, &num_cpu);
	    if (err == CL_SUCCESS) clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_GPU, 0, NULL, &num_gpu);
	    if (err == CL_SUCCESS) clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_ACCELERATOR, 0, NULL, &num_accell);
	    if (err == CL_SUCCESS) {
		cl_uint j, units, clock;
		cl_ulong mem;
		cl_device_id devices[HST_MAX_OPENCL_DEVICES]; 

		pyhst_info(" Platform %u: %s, %s (CPU: %u, GPU: %u, Other: %u)", i, name, version, num_cpu, num_gpu, num_accell );

		err = clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_ALL, HST_MAX_OPENCL_DEVICES, devices, &num_devices);
		for (j = 0; j < num_devices; ++j) {
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DEVICE_NAME, sizeof(name), name, NULL);
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DRIVER_VERSION, sizeof(version), version, NULL);
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(units), &units, NULL);
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(clock), &clock, NULL);
		    if (err == CL_SUCCESS) err = clGetDeviceInfo(devices[j], CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(mem), &mem, NULL);
    		    if (err == CL_SUCCESS) pyhst_debug(" * Device %u: %s v. %s: %u SM, %u MHZ, %lu MB", j, name, version, units, clock, mem/1024/1024);
		    else pyhst_debug(" * Device %u: Failed", i);
		}
	    } else pyhst_info(" Platform %u: Failed", i);
	}
    }

	// Looking for GPU-enabled platform first
    if (flags&HST_RECONSTRUCTOR_USE_GPU) {
	    // FIXME: We should exclude slow devices (check memory, capabilities, etc.)
	for (i = 0; i < num_platforms; i++) {
	    err = clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_GPU, 0, NULL, &num_devices);
	    if ((err == CL_SUCCESS)&&(num_devices > max)) {
		max = num_devices;
		selected_platform = platforms[i];
	    }
	}
    }

    if (!max) {
#ifdef HW_FALLBACK_CPU
	flags|=HST_RECONSTRUCTOR_USE_CPU;
#endif /* HW_FALLBACK_CPU */

	if (flags&HST_RECONSTRUCTOR_USE_CPU) {
	    for (i = 0; i < num_platforms; i++) {
		err = clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_CPU, 0, NULL, &num_devices);
		if ((err == CL_SUCCESS)&&(num_devices > 0)) {
		    max = num_devices;
		    selected_platform = platforms[i];
		    break;
		}
	    }
	}
    }
    
    if (max) {
        err = clGetPlatformInfo (selected_platform, CL_PLATFORM_VENDOR, sizeof(name), name, NULL);
        if (err ==  CL_SUCCESS) {
    	    char *found;

	    pyhst_warning("OpenCL Platform: %s", name);
	    
	    found = strstr(name, "ATI");
    	    if (!found) found = strstr(name, "AMD");
    	    if (found) {
    		switch (found[3]) {
    		    case ' ':
    		    case '\t':
    		    case 0:
    			amd_platform = 1;
    		}
    	    } else {
    		found = strcasestr(name, "Advanced Micro Devices");
    		if (found) amd_platform = 1;
    	    }
	    
	    found = strcasestr(name, "Intel");
	    if (found) intel_platform = 1;

	    found = strcasestr(name, "NVIDIA");
	    if (found) nvidia_platform = 1;
    	}
	
	if (nvidia_platform) {
	    have_opencl11 = 0; 	// There is a beta driver, but it is incompatible with CUDA 4.0. Very nice of them!
	}
    
	g_hst_platform.platform_id = selected_platform;
	clGetDeviceIDs(g_hst_platform.platform_id,
                ((flags&HST_RECONSTRUCTOR_USE_CPU)?CL_DEVICE_TYPE_CPU:0)|((flags&HST_RECONSTRUCTOR_USE_GPU)?CL_DEVICE_TYPE_GPU:0),
                HST_MAX_OPENCL_DEVICES,
                g_hst_platform.device_ids,
                &(g_hst_platform.num_devices));
	

	int move = 0;
	int have_gpu = 0;
	for (i = 0; i < g_hst_platform.num_devices; i++) {
	    cl_device_type device_type;
	    cl_bool device_avail;
	    cl_ulong device_mem;
		
	    if ((amd_platform)&&(flags&HST_RECONSTRUCTOR_USE_GPU)) {
		err = clGetDeviceInfo(g_hst_platform.device_ids[i], CL_DEVICE_TYPE, sizeof(cl_device_type), &device_type, NULL);
		CL_CHECK_ERROR(err);
	    }

	    err = clGetDeviceInfo(g_hst_platform.device_ids[i], CL_DEVICE_AVAILABLE, sizeof(cl_bool), &device_avail, NULL);
	    CL_CHECK_ERROR(err);

#ifdef HST_CHECK_MEMORY
	    err = clGetDeviceInfo(g_hst_platform.device_ids[i], CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(cl_ulong), &device_mem, NULL);
	    CL_CHECK_ERROR(err);
#endif /* HST_CHECK_MEMORY */

	    if (move) {
		g_hst_platform.device_ids[i - move] =  g_hst_platform.device_ids[i];
	    }
		
	    if (!device_avail) ++move;
#ifdef HST_CHECK_MEMORY
	    else if (device_mem < HST_CHECK_MEMORY) ++move;
#endif /* HST_CHECK_MEMORY */
	    else if ((amd_platform)&&(device_type == CL_DEVICE_TYPE_GPU)) {
		if (have_gpu) ++move;
		else have_gpu = 1;
	    }

	}
	
	g_hst_platform.num_devices -= move;
    } else {
	g_hst_platform.num_devices = 0;
    }

    pyhst_info("OpenCL Devices: %u\n", g_hst_platform.num_devices);
    
    hst_opencl_info.devices = g_hst_platform.num_devices;
    return &hst_opencl_info;
}

void hst_opencl_free()
{
}


