/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define BLOCK_SIZE 16

const sampler_t volumeSampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP | CLK_FILTER_LINEAR; 


__kernel void hst_filter(__global float *data, __global float *filter)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int id = idy*get_global_size(0) + idx;
    data[id] = data[id] * filter[idx];
}

__kernel void hst_normalize_idft(__global float *data)
{
    const int idx = get_global_id(0);
    const int dim_fft = get_global_size(0);
    data[2*idx] = data[2*idx] / dim_fft;
}

__kernel void hst_spread(__global float *out, __global float *in, const int num_bins)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int dpitch = get_global_size(0)*2;

    /* May diverge but not possible to reduce latency, because num_bins can 
       be arbitrary and not be aligned. */
    if (idx >= num_bins) {
        out[idy*dpitch + idx*2] = 0.0f;
        out[idy*dpitch + idx*2 + 1] = 0.0f;
    }
    else {
        out[idy*dpitch + idx*2] = in[idy*num_bins + idx];
        out[idy*dpitch + idx*2 + 1] = 0.0f;
    }
}

__kernel void hst_pack(__global float *in, __global float *out, const int num_bins)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int dpitch = get_global_size(0)*2;

    if (idx < num_bins) {
        out[idy*num_bins + idx] = in[idy*dpitch + idx*2];
    }
}

__kernel void hst_backproject_cpu(const int num_proj,
    const int num_bins,
    const float off_x,
    const float off_y,
    __global float *sinogram,
    __global float *slice,
    __constant float4 *all_table)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int slice_width = get_global_size(0);

    // TODO: maybe use float4 for optimal vectorization on Intel CPUs
    float2 b = (float2) (idx + off_x, idy + off_y);
    float corr;
    float sum = 0.0f;

//#pragma unroll 8
    for(int proj = 0; proj < num_proj; proj++) {
        float2 s = (float2) (all_table[proj].x, all_table[proj].y);
        s = s*b;
        corr = all_table[proj].z;
        sum += sinogram[(int)(proj*num_bins + corr + s.x - s.y)];
    }
    slice[idy*slice_width + idx] = sum;
}


__kernel void hst_backproject_gpu_nearest(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all,
    __local float *shared) {

    __local float *cache = shared;
    __local float *cache_subh = shared + 48 * BLOCK_SIZE;			// cache_subh[BLOCK_SIZE_X][2 * BLOCK_SIZE_Y + 1]
    __local float *cache_sin = cache_subh +  BLOCK_SIZE * (2*BLOCK_SIZE + 1);	// cache_sin[2 * BLOCK_SIZE_Y]
/*
    __local float cache[2 * BLOCK_SIZE * 24];
    __local float cache_subh[BLOCK_SIZE * (2 * BLOCK_SIZE + 1)];
    __local float cache_sin[2 * BLOCK_SIZE];
*/
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = get_group_id(0) * BLOCK_SIZE;
    const int bidy = get_group_id(1) * BLOCK_SIZE;

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const float x = bx + tidx;
    const float y = by + tidy;
    
    const int slice_width = get_global_size(0);

    float res = 0;
    float4 all1, all2;

    int minh1, minh2;

    const int prj = tidx >> 3;
    const int pos = BLOCK_SIZE + (tidx & 7);

    const int tidy2 = tidy << 1;
    const int tidy21 = tidy2 + 1;
    const int tidy2p = tidy2 + prj;

    for (int proje=0; proje<num_proj; proje += 2 * BLOCK_SIZE) {
	int proj = proje + tidy2;
	
        all1 = c_all[proj];
        all2 = c_all[proj + 1];
	minh1 = (int)(all1.z + bx * all1.x - by * all1.y + all1.w);
	minh2 = (int)(all2.z + bx * all2.x - by * all2.y + all2.w);
		
	cache_subh[tidx * (2 * BLOCK_SIZE + 1) + tidy2] = all1.z + x * all1.x - minh1;
	cache_subh[tidx * (2 * BLOCK_SIZE + 1) + tidy21] = all2.z + x * all2.x - minh2;
	
	if (tidx == 0) {
	    cache_sin[tidy2] = all1.y;
	    cache_sin[tidy21] = all2.y;
	}

        cache[tidy2 * 24 + tidx] = read_imagef(sinogram, volumeSampler, (float2)(minh1 + tidx, proj + 0.5f)).x;
        cache[tidy21 * 24 + tidx] = read_imagef(sinogram, volumeSampler, (float2)(minh2 + tidx, proj + 1.5f)).x;
        cache[tidy2p * 24 + pos] = read_imagef(sinogram, volumeSampler, (float2)((prj?minh2:minh1) + pos, proj + prj + 0.5f)).x;

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
	    int j = 2 * i;
	    minh1 = cache_subh[tidx * (2 * BLOCK_SIZE + 1) + j] - y * cache_sin[j];
	    res += cache[j * 24 + minh1];

	    ++j;
	    minh1 = cache_subh[tidx * (2 * BLOCK_SIZE + 1) + j] - y * cache_sin[j];
	    res += cache[j * 24 + minh1];
	}

        barrier(CLK_LOCAL_MEM_FENCE);

    }

    slice[idy*slice_width + idx] = res;
}


__kernel void hst_backproject_gpu_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][6 * BLOCK_SIZE_X];
    __local float *cache_subh = cache + 6 * BLOCK_SIZE * BLOCK_SIZE;		// cache_subh[BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __local float *cache_sin = cache_subh + BLOCK_SIZE * (BLOCK_SIZE + 1);	// cache_sin[BLOCK_SIZE_Y]

    const int idx = get_global_id(0);
    const int idy = get_global_id(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = get_group_id(0) * BLOCK_SIZE;
    const int bidy = get_group_id(1) * BLOCK_SIZE;

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const float x = bx + tidx;
    const float y = by + tidy;
    
    const int slice_width = get_global_size(0);

    float res = 0;
    
//#pragma unroll 8
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	int minh = (int)(all.z + bx * all.x - by * all.y + all.w);

	cache_subh[tidx * (BLOCK_SIZE + 1) + tidy] = 4 * (all.z + x * all.x - minh);
	if (tidx == 0) cache_sin[tidy] = 4 * all.y;

#pragma unroll 6
	for (int i = 0; i < 6; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * 6 * BLOCK_SIZE + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + ((float)pos)/4, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
	    minh = cache_subh[tidx * (BLOCK_SIZE + 1) + i] - y * cache_sin[i];
	    res += cache[i * 6 * BLOCK_SIZE + minh];
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[idy*slice_width + idx] = res;
}

__kernel void hst_backproject_gpu_linear(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float2 *cache = (__local float2*)shared;						// cache[2 * BLOCK_SIZE_Y][24];
    __local float *cache_subh = shared + 96 * BLOCK_SIZE;					// cache_subh[BLOCK_SIZE_X][2 * BLOCK_SIZE_Y + 1]
    __local float *cache_sin = cache_subh +  BLOCK_SIZE * (2*BLOCK_SIZE + 1);			// cache_sin[2 * BLOCK_SIZE_Y]


    const int idx = get_global_id(0);
    const int idy = get_global_id(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = get_group_id(0) * BLOCK_SIZE;
    const int bidy = get_group_id(1) * BLOCK_SIZE;

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const float x = bx + tidx;
    const float y = by + tidy;
    
    const int slice_width = get_global_size(0);

    float res = 0;
    float4 all1, all2;

    float minh;
    int minh1, minh2;

    const int prj = tidx >> 3;
    const int pos = BLOCK_SIZE + (tidx & 7);

    const int tidy2 = tidy << 1;
    const int tidy21 = tidy2 + 1;
    const int tidy2p = tidy2 + prj;

    for (int proje=0; proje<num_proj; proje += 2 * BLOCK_SIZE) {
	int proj = proje + tidy2;
	
        all1 = c_all[proj];
        all2 = c_all[proj + 1];
	minh1 = (int)(all1.z + bx * all1.x - by * all1.y + all1.w);
	minh2 = (int)(all2.z + bx * all2.x - by * all2.y + all2.w);
		
	cache_subh[tidx * (2 * BLOCK_SIZE + 1) + tidy2] = all1.z + x * all1.x - minh1;
	cache_subh[tidx * (2 * BLOCK_SIZE + 1) + tidy21] = all2.z + x * all2.x - minh2;
	
	if (tidx == 0) {
	    cache_sin[tidy2] = all1.y;
	    cache_sin[tidy21] = all2.y;
	}

        cache[tidy2 * 24 + tidx].x = read_imagef(sinogram, volumeSampler, (float2)(minh1 + tidx, proj + 0.5f)).x;
        cache[tidy21 * 24 + tidx].x = read_imagef(sinogram, volumeSampler, (float2)(minh2 + tidx, proj + 1.5f)).x;
        cache[tidy2p * 24 + pos].x = read_imagef(sinogram, volumeSampler, (float2)((prj?minh2:minh1) + pos, proj + prj + 0.5f)).x;

        barrier(CLK_LOCAL_MEM_FENCE);

    	cache[tidy2 * 24 + tidx].y = cache[tidy2 * 24 + tidx + 1].x - cache[tidy2 * 24 + tidx].x;
	cache[tidy21 * 24 + tidx].y = cache[tidy21 * 24 + tidx + 1].x - cache[tidy21 * 24 + tidx].x;
	if (pos < 23) {
	    cache[tidy2p * 24 + pos].y = cache[tidy2p * 24 + pos + 1].x - cache[tidy2p * 24 + pos].x;
	} else {
	    cache[tidy2p * 24 + pos].y = 0;
	}

        barrier(CLK_LOCAL_MEM_FENCE);
	
#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
	    int j = 2 * i;
	    float2 v;
	    
	    minh = cache_subh[tidx * (2 * BLOCK_SIZE + 1) + j] - y * cache_sin[j];
	    minh1 = (int)minh;
	    v = cache[j * 24 + minh1];
	    res += v.x + (minh - minh1) * v.y;

	    ++j;
	    minh = cache_subh[tidx * (2 * BLOCK_SIZE + 1) + j] - y * cache_sin[j];
	    minh1 = (int)minh;
	    v = cache[j * 24 + minh1];
	    res += v.x + (minh - minh1) * v.y;
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[idy*slice_width + idx] = res;
}



/*
__kernel void hst_pack(__global float *in, __write_only image2d_t out, const int num_bins)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int dpitch = get_global_size(0)*2;

    if (idx < num_bins) {
//        out[idy*num_bins + idx] = in[idy*dpitch + idx*2];
	int2 pos = {idx, idy};
	float4 val = {in[idy*dpitch + idx*2], 0, 0, 0};
	write_imagef(out, pos, val);
    }
}
*/


__kernel void hst_backproject_gpu(const int num_proj,
    const int num_bins,
    const float off_x,
    const float off_y,
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *all_table)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int slice_width = get_global_size(0);

    float h;
    float bx = idx + off_x;
    float by = -(idy + off_y);
    float sum = 0.0f;

    for(int proj = 0; proj < num_proj; proj++) {
	h = mad(by, all_table[proj].y, mad(bx, all_table[proj].x, all_table[proj].z)); 
	sum += read_imagef(sinogram, volumeSampler, (float2)(h, proj + 0.5f)).x;
    }
    
    slice[idy*slice_width + idx] = sum;
}


//HST_OPENCL_CONST_HACK - do not edit, the string is used to remove following code if hack is disabled

__kernel void hst_backproject_gpu_const(const int num_proj,
    const int num_bins,
    const float off_x,
    const float off_y,
    __read_only image2d_t sinogram,
    __global float *slice)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int slice_width = get_global_size(0);

    float h;
    float bx = idx + off_x;
    float by = -(idy + off_y);
    float sum = 0.0f;

    for(int proj = 0; proj < num_proj; proj++) {
	h = mad(by, as_float(sin_itable[proj]), mad(bx, as_float(cos_itable[proj]), as_float(axis_itable[proj]))); 
	sum += read_imagef(sinogram, volumeSampler, (float2)(h, proj + 0.5f)).x;
    }
    slice[idy*slice_width + idx] = sum;
}
